package freenet.client.async;

import com.db4o.ObjectContainer;

import freenet.client.FECCallback;
import freenet.client.FECCodec;
import freenet.client.FECJob;
import freenet.client.FECQueue;
import freenet.client.SplitfileBlock;
import freenet.keys.CHKBlock;
import freenet.support.LogThresholdCallback;
import freenet.support.Logger;
import freenet.support.Logger.LogLevel;
import freenet.support.api.Bucket;

public class SplitFileInserterCrossSegment implements FECCallback {
	
	private final boolean persistent;
	private final int dataBlocks;
	private final int crossCheckBlocks;
	private final SplitFileInserterSegment[] segments;
	private final int[] blockNumbers;
	private final ClientRequester parent;
	private final SplitFileInserter parentInserter;
	private final int segNum;
	
	final short splitfileType;
	
	private transient int counter;
	
	private transient FECCodec codec;
	
	private static volatile boolean logMINOR;

	static {
		Logger.registerLogThresholdCallback(new LogThresholdCallback() {
			@Override
			public void shouldUpdate() {
				logMINOR = Logger.shouldLog(LogLevel.MINOR, SplitFileInserterCrossSegment.class);
			}
		});
	}
	
	SplitFileInserterCrossSegment(boolean persistent, int dataBlocks, int crossCheckBlocks, ClientRequester parent, short splitfileType, SplitFileInserter parentInserter, int segNum) {
		this.persistent = persistent;
		this.dataBlocks = dataBlocks;
		this.crossCheckBlocks = crossCheckBlocks;
		int totalBlocks = dataBlocks + crossCheckBlocks;
		this.segments = new SplitFileInserterSegment[totalBlocks];
		this.blockNumbers = new int[totalBlocks];
		this.parent = parent;
		this.splitfileType = splitfileType;
		this.parentInserter = parentInserter;
		this.segNum = segNum;
	}
	
	public void addDataBlock(SplitFileInserterSegment seg, int blockNum) {
		if(logMINOR) Logger.minor(this, "Allocated "+counter+" blocks: block "+blockNum+" on segment "+seg.segNo);
		segments[counter] = seg;
		blockNumbers[counter] = blockNum;
		counter++;
	}

	/** Start the encode */
	public void start(ObjectContainer container, ClientContext context) {
		if(logMINOR) Logger.minor(this, "Scheduling encode for cross segment");
		// Schedule encode job.
		SplitfileBlock[] decodeData = new SplitfileBlock[dataBlocks];
		SplitfileBlock[] decodeCheck = new SplitfileBlock[segments.length - dataBlocks];
		for(int i=0;i<decodeData.length;i++) {
			MinimalSplitfileBlock wrapper = new MinimalSplitfileBlock(i);
			SplitFileInserterSegment seg = segments[i];
			boolean active = true;
			if(persistent) {
				active = container.ext().isActive(seg);
				if(!active) container.activate(seg, 1);
			}
			Bucket data = seg.getBucket(blockNumbers[i]);
			wrapper.assertSetData(data);
			if(persistent) container.activate(data, Integer.MAX_VALUE);
			if(!active) container.deactivate(seg, 1);
			decodeData[i] = wrapper;
		}
		for(int i=0;i<decodeCheck.length;i++) {
			decodeCheck[i] = new MinimalSplitfileBlock(i);
		}
		FECQueue queue = context.fecQueue;
		if(codec == null)
			codec = FECCodec.getCodec(splitfileType, dataBlocks, decodeCheck.length);
		FECJob job = new FECJob(codec, queue, decodeData, decodeCheck, CHKBlock.DATA_LENGTH, context.getBucketFactory(persistent), this, false, getPriorityClass(container), persistent);
		codec.addToQueue(job, 
				queue, container);
	}
	
	private short getPriorityClass(ObjectContainer container) {
		boolean parentActive = true;
		if(persistent) {
			parentActive = container.ext().isActive(parent);
			if(!parentActive) container.activate(parent, 1);
		}
		short ret = parent.getPriorityClass();
		if(!parentActive) container.deactivate(parent, 1);
		return ret;
	}

	public void storeTo(ObjectContainer container) {
		container.store(this);
	}

	@Override
	public void onDecodedSegment(ObjectContainer container, ClientContext context, FECJob job, Bucket[] dataBuckets, Bucket[] checkBuckets, SplitfileBlock[] dataBlocks, SplitfileBlock[] checkBlocks) {
		// Can't happen.
		throw new UnsupportedOperationException();
	}

	@Override
	public void onEncodedSegment(ObjectContainer container, ClientContext context, FECJob job, Bucket[] dataBuckets, Bucket[] checkBuckets, SplitfileBlock[] dataBlocks, SplitfileBlock[] checkBlocks) {
		for(int i=0;i<crossCheckBlocks;i++) {
			SplitFileInserterSegment seg = segments[i + this.dataBlocks];
			int blockNum = blockNumbers[i + this.dataBlocks];
			if(persistent) container.activate(seg, 1);
			seg.onEncodedCrossCheckBlock(blockNum, checkBlocks[i].getData(), container, context);
			if(persistent) container.deactivate(seg, 1);
			checkBlocks[i].clearData();
			if(persistent) container.delete(checkBlocks[i]);
		}
		for(int i=0;i<dataBlocks.length;i++) {
			dataBlocks[i].clearData();
			if(persistent) container.delete(dataBlocks[i]);
		}
		if(logMINOR) Logger.minor(this, "Completed encode for cross segment");
		if(persistent) container.activate(parentInserter, 1);
		parentInserter.clearCrossSegment(segNum, this, container, context);
		if(persistent) container.deactivate(parentInserter, 1);
		if(persistent) removeFrom(container);
	}

	private void removeFrom(ObjectContainer container) {
		container.delete(this);
	}

	@Override
	public void onFailed(Throwable t, ObjectContainer container, ClientContext context) {
		Logger.error(this, "Encode or decode failed for cross segment: "+this, t);
	}

}
